#lang racket/base

(require racket/match
         "lc-ast.rkt"
         "lc-parser.rkt")

;; Church encoding for booleans
(define ctrue (lc-parser "λ t. (λ f. (t NOPE))"))
(define cfalse (lc-parser "λ t. (λ f. (f NOPE))"))

;; Church encoding of numbers
(define (apply-n f n x)
  (if (zero? n)
      x
      (app f (apply-n f (- n 1) x))))
(define (church-num n)
  (if (zero? n)
      (abs 'f (abs 'z (var 'z)))
      (abs 'f (abs 'z (apply-n (var 'f) n (var 'z))))))

;; Operations on numbers
(define csucc (lc-parser "(λ n. (λ f. (λ z. (f (n f z)))))"))
(define csum (lc-parser "(λ n. (λ m. (λ f. (λ z. ((m f) (n f z))))))"))
(define cmul (lc-parser "(λ n. (λ m. (λ f. (λ z. ((m (n f)) z)))))"))
(define cpred (lc-parser "(λ n. (λ f. (λ z. (((n (λ g. (λ h. (h (g f))))) (λ u. z)) (λ u. u)))))"))
(define csub (abs 'n (abs 'm (app (app (var 'm) cpred) (var 'n)))))
(define czerop (abs 'n (app (app (var 'n) (abs '_ cfalse)) ctrue)))

;; Church encoding of lists
(define ccons (lc-parser "(λ a. (λ b. (λ c. (λ n. (c a b)))))"))
(define cnil (lc-parser "(λ c. (λ n. (n NOPE)))"))

;; Operations on lists
(define ccar (lc-parser "(λ p. (p (λ a. (λ b. a)) error))"))
(define ccdr (lc-parser "(λ p. (p (λ a. (λ b. b)) error))"))
(define cnullp (abs 'p (app (app (var 'p) (abs 'a (abs 'b cfalse))) (abs 'n ctrue))))
(define cpairp (abs 'p (app (app (var 'p) (abs 'a (abs 'b ctrue))) (abs 'n cfalse))))

;; Y combinator
(define Y (lc-parser "((λ y. (λ F. (F (λ x. (y y F x))))) (λ y. (λ F. (F (λ x. (y y F x))))))"))

;; Compile function
(define (compile expr)

  (match expr
    (#t ctrue)
    (#f cfalse)
    (`(if ,test ,yes ,no)
     (compile `(,test (lambda () ,yes) (lambda () ,no))))
    (`(and ,a ,b)
     (compile `(if ,a ,b #f)))
    (`(or ,a ,b)
     (compile `(if ,a #t ,b)))
    (`(not ,a)
     (compile `(if ,a #f #t)))

    ((? (λ (v) (and (number? v) (or (zero? v) (exact-positive-integer? v)))) v)
     (church-num v))
    (`(+ ,x ,y)
     (app (app csum (compile x)) (compile y)))
    (`(- ,x ,y)
     (app (app csub (compile x)) (compile y)))
    (`(pred ,x)
     (app cpred (compile x)))
    (`(succ ,x)
     (app csucc (compile x)))
    (`(* ,x ,y)
     (app (app cmul (compile x)) (compile y)))
    (`(zero? ,x)
     (app czerop (compile x)))

    ((or '(list) '() 'nil)
     cnil)
    (`(list ,a)
     (compile `(cons ,a nil)))
    (`(list ,a . ,r)
     (compile `(cons ,a (list ,@r))))
    (`(cons ,a ,b)
     (app (app ccons (compile a)) (compile b)))
    (`(car ,l)
     (app ccar (compile l)))
    (`(cdr ,l)
     (app ccdr (compile l)))
    (`(null? ,l)
     (app cnullp (compile l)))
    (`(pair? ,l)
     (app cpairp (compile l)))

    (`(lambda () ,body)
     (abs '_ (compile body)))
    (`(lambda (,arg) ,body)
     (abs arg (compile body)))
    (`(lambda (,arg . ,args) ,body)
     (abs arg (compile `(lambda ,args ,body))))

    (`(let ((,var ,val) ...) ,body)
     (compile `((lambda (,@var) ,body) ,@val)))
    (`(letrec ((,var ,val)) ,body)
     (compile `(let ((,var (,Y (lambda (,var) ,val)))) ,body)))

    (`(,fun)
     (app (compile fun) (var 'NOPE)))
    (`(,fun ,arg)
     (app (compile fun) (compile arg)))
    (`(,fun ,arg . ,args)
     (compile `((,fun ,arg) ,@args)))

    ((? symbol? v) (var v))

    ((? Term? t) t)

    (else
     (begin
       (eprintf "Error: this should not happen!\nThe bugging expr is '~a'.\n"
                expr)
       (exit 1)))))

;; Transform λ-calculus AST to quoted Racket code
(define (to-racket term)
  (match term
    ((Var n i) i)
    ((Abs v i b) `(λ (,i) ,(to-racket b)))
    ((App f (Var 'NOPE _)) `(,(to-racket f) '_))
    ((App f a) `(,(to-racket f) ,(to-racket a)))))

;; Transform the result back to native values
(define (lc-list->rkt-list lc->rkt l)
  ((l (λ (head)
        (λ (tail)
          (cons (lc->rkt head)
                (lc-list->rkt-list lc->rkt tail)))))
   (λ (_) '())))
(define (lc-num->rkt-num n) ((n add1) 0))
(define (lc-bool->rkt-bool b) ((b (λ (_) #t)) (λ (_) #f)))

;; Build the function that will transform the result back to a native value
;; with respect to its type
(define (lc->rkt rt)
  (match rt
    ('n lc-num->rkt-num)
    ('b lc-bool->rkt-bool)
    (`(l ,rt) (lambda (res) (lc-list->rkt-list (lc->rkt rt) res)))))

;; Main program
(define argv (current-command-line-arguments))
(cond
  ((= (vector-length argv) 2)
   (define in0 (open-input-string (vector-ref argv 0)))
   (define rt (read in0))
   (close-input-port in0)
   (define in1 (open-input-string (vector-ref argv 1)))
   (define code (read in1))
   (close-input-port in1)
   (displayln "The code is:")
   (displayln code)
   (newline)
   (define term (compile code))
   (displayln "The same code compiled to λ-calculus is:")
   (lc-pp term)
   (newline)
   (displayln "Once evaluated, and transformed back to native values it gives:")
   (displayln ((lc->rkt rt) (eval (to-racket term) (make-base-namespace))))
   (exit 0))
  (else
   (eprintf "Usage: racket compile2lc.rkt \"return type\" \"code\"\n")
   (eprintf "  return type syntax is:\n")
   (eprintf "  RT ::= \"n\"               (for number)\n")
   (eprintf "       | \"b\"               (for boolean)\n")
   (eprintf "       | \"(l \" RT \")\"      (for list of RT)\n")
   (exit 1)))
